"""
Implements optimisation-based methods for solving linear inverse problems.

This module provides classical, deterministic approaches to inversion that seek
a single "best-fit" model. These methods are typically formulated as finding
the model `u` that minimizes a cost functional.

Key Classes
-----------
- `LinearLeastSquaresInversion`: Solves the inverse problem by minimizing a
  Tikhonov-regularized least-squares functional.
- `LinearMinimumNormInversion`: Finds the model with the smallest norm that
  fits the data to a statistically acceptable degree using the discrepancy
  principle.
- `ConstrainedLinearLeastSquaresInversion`: Solves a linear inverse problem
  subject to an affine subspace constraint.
"""

from __future__ import annotations
from typing import Optional, Union

from .nonlinear_operators import NonLinearOperator
from .affine_operators import AffineOperator
from .inversion import LinearInversion
from .forward_problem import LinearForwardProblem
from .linear_operators import LinearOperator
from .linear_solvers import LinearSolver, IterativeLinearSolver
from .hilbert_space import Vector
from .subspaces import AffineSubspace


class LinearLeastSquaresInversion(LinearInversion):
    """
    Solves a linear inverse problem using Tikhonov-regularized least-squares.

    This method finds the model `u` that minimizes the functional:
    `J(u) = ||A(u) - d||² + α² * ||u||²`
    """

    def __init__(self, forward_problem: LinearForwardProblem, /) -> None:
        super().__init__(forward_problem)
        if self.forward_problem.data_error_measure_set:
            self.assert_inverse_data_covariance()

    def normal_operator(self, damping: float) -> LinearOperator:
        """Returns the Tikhonov-regularized normal operator (A*WA + αI)."""
        if damping < 0:
            raise ValueError("Damping parameter must be non-negative.")

        forward_operator = self.forward_problem.forward_operator
        identity = self.forward_problem.model_space.identity_operator()

        if self.forward_problem.data_error_measure_set:
            inverse_data_covariance = (
                self.forward_problem.data_error_measure.inverse_covariance
            )
            return (
                forward_operator.adjoint @ inverse_data_covariance @ forward_operator
                + damping * identity
            )
        else:
            return forward_operator.adjoint @ forward_operator + damping * identity

    def normal_rhs(self, data: Vector) -> Vector:
        """Returns the right hand side of the normal equations (A*W d)."""
        forward_operator = self.forward_problem.forward_operator

        if self.forward_problem.data_error_measure_set:
            inverse_data_covariance = (
                self.forward_problem.data_error_measure.inverse_covariance
            )
            shifted_data = self.forward_problem.data_space.subtract(
                data, self.forward_problem.data_error_measure.expectation
            )
            return (forward_operator.adjoint @ inverse_data_covariance)(shifted_data)
        else:
            return forward_operator.adjoint(data)

    def least_squares_operator(
        self,
        damping: float,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
    ) -> Union[LinearOperator, AffineOperator]:
        """
        Returns an operator that maps data to the least-squares solution.

        Args:
            damping: The Tikhonov damping parameter, alpha.
            solver: The linear solver for inverting the normal operator.
            preconditioner: Either a direct LinearOperator or a LinearSolver
                method (factory) used to generate the preconditioner.
        """
        forward_operator = self.forward_problem.forward_operator
        normal_operator = self.normal_operator(damping)

        resolved_preconditioner = None
        if preconditioner is not None:
            if isinstance(preconditioner, LinearOperator):
                resolved_preconditioner = preconditioner
            elif isinstance(preconditioner, LinearSolver):
                resolved_preconditioner = preconditioner(normal_operator)
            else:
                raise TypeError(
                    "Preconditioner must be a LinearOperator or LinearSolver."
                )

        if isinstance(solver, IterativeLinearSolver):
            inverse_normal_operator = solver(
                normal_operator, preconditioner=resolved_preconditioner
            )
        else:
            inverse_normal_operator = solver(normal_operator)

        if self.forward_problem.data_error_measure_set:

            inverse_data_covariance = (
                self.forward_problem.data_error_measure.inverse_covariance
            )

            linear_part = (
                inverse_normal_operator
                @ forward_operator.adjoint
                @ inverse_data_covariance
            )

            expected_data = self.forward_problem.data_error_measure.expectation
            translation = self.model_space.negative(linear_part(expected_data))

            return AffineOperator(linear_part, translation)
        else:
            return inverse_normal_operator @ forward_operator.adjoint


class ConstrainedLinearLeastSquaresInversion(LinearInversion):
    """Solves a linear inverse problem subject to an affine subspace constraint."""

    def __init__(
        self, forward_problem: LinearForwardProblem, constraint: AffineSubspace
    ) -> None:
        super().__init__(forward_problem)
        self._constraint = constraint

        # Use the projection operator to seamlessly grab the base translation
        proj_op = constraint.projection_operator
        self._u_base = proj_op.translation_part

        # The reduced operator is just A @ P
        reduced_operator = forward_problem.forward_operator @ proj_op.linear_part

        self._reduced_forward_problem = LinearForwardProblem(
            reduced_operator,
            data_error_measure=(
                forward_problem.data_error_measure
                if forward_problem.data_error_measure_set
                else None
            ),
        )

        self._unconstrained_inversion = LinearLeastSquaresInversion(
            self._reduced_forward_problem
        )

    def least_squares_operator(
        self,
        damping: float,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
    ) -> AffineOperator:
        """Maps data to the constrained least-squares solution."""
        reduced_op = self._unconstrained_inversion.least_squares_operator(
            damping, solver, preconditioner=preconditioner
        )
        data_offset = self.forward_problem.forward_operator(self._u_base)
        domain = self.data_space
        codomain = self.model_space
        shift_in = AffineOperator(
            domain.identity_operator(), domain.negative(data_offset)
        )
        shift_out = AffineOperator(codomain.identity_operator(), self._u_base)
        return shift_out @ reduced_op @ shift_in


class LinearMinimumNormInversion(LinearInversion):
    """Finds a regularized solution using the discrepancy principle."""

    def __init__(self, forward_problem: LinearForwardProblem) -> None:
        super().__init__(forward_problem)
        if self.forward_problem.data_error_measure_set:
            self.assert_inverse_data_covariance()

    def minimum_norm_operator(
        self,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
        significance_level: float = 0.95,
        minimum_damping: float = 0.0,
        maxiter: int = 100,
        rtol: float = 1.0e-6,
        atol: float = 0.0,
    ) -> Union[NonLinearOperator, LinearOperator]:
        """
        Maps data to the minimum-norm solution matching target chi-squared.

        The returned NonLinearOperator includes the exact analytical Fréchet
        derivative of the discrepancy search, complete with its adjoint mapping.
        """
        if self.forward_problem.data_error_measure_set:
            critical_value = self.forward_problem.critical_chi_squared(
                significance_level
            )
            lsq_inversion = LinearLeastSquaresInversion(self.forward_problem)

            def get_model_for_damping(
                damping: float, data: Vector, model0: Optional[Vector] = None
            ) -> tuple[Vector, float]:
                normal_operator = lsq_inversion.normal_operator(damping)
                normal_rhs = lsq_inversion.normal_rhs(data)

                res_precond = None
                if preconditioner is not None:
                    if isinstance(preconditioner, LinearOperator):
                        res_precond = preconditioner
                    else:
                        res_precond = preconditioner(normal_operator)

                if isinstance(solver, IterativeLinearSolver):
                    model = solver.solve_linear_system(
                        normal_operator, res_precond, normal_rhs, model0
                    )
                else:
                    inverse_normal_operator = solver(normal_operator)
                    model = inverse_normal_operator(normal_rhs)

                chi_squared = self.forward_problem.chi_squared(model, data)
                return model, chi_squared

            def _solve_discrepancy(data: Vector) -> tuple[Vector, float]:
                """Internal helper to find the model and optimal damping."""
                chi_squared = self.forward_problem.chi_squared_from_residual(data)

                if chi_squared <= critical_value:
                    return self.model_space.zero, 0.0

                damping = 1.0
                _, chi_squared = get_model_for_damping(damping, data)
                damping_lower = damping if chi_squared <= critical_value else None
                damping_upper = damping if chi_squared > critical_value else None

                it = 0
                if damping_lower is None:
                    while chi_squared > critical_value and it < maxiter:
                        it += 1
                        damping /= 2.0
                        _, chi_squared = get_model_for_damping(damping, data)
                        if damping < minimum_damping:
                            raise RuntimeError("Discrepancy principle failed.")
                    damping_lower = damping

                it = 0
                if damping_upper is None:
                    while chi_squared < critical_value and it < maxiter:
                        it += 1
                        damping *= 2.0
                        _, chi_squared = get_model_for_damping(damping, data)
                    damping_upper = damping

                model0 = None
                for _ in range(maxiter):
                    damping = 0.5 * (damping_lower + damping_upper)
                    model, chi_squared = get_model_for_damping(damping, data, model0)

                    if chi_squared < critical_value:
                        damping_lower = damping
                    else:
                        damping_upper = damping

                    if damping_upper - damping_lower < atol + rtol * (
                        damping_lower + damping_upper
                    ):
                        return model, damping
                    model0 = model

                raise RuntimeError("Bracketing search failed to converge.")

            def mapping(data: Vector) -> Vector:
                model, _ = _solve_discrepancy(data)
                return model

            def derivative(data: Vector) -> LinearOperator:
                model, damping = _solve_discrepancy(data)

                if damping == 0.0:
                    return self.data_space.zero_operator(self.model_space)

                # 1. Standard Tikhonov linear part (L)
                lsq_op = lsq_inversion.least_squares_operator(
                    damping, solver, preconditioner=preconditioner
                )
                linear_part = lsq_op.linear_part

                # 2. Reconstruct H^{-1} for the denominator term
                normal_operator = lsq_inversion.normal_operator(damping)
                res_precond = None
                if preconditioner is not None:
                    if isinstance(preconditioner, LinearOperator):
                        res_precond = preconditioner
                    else:
                        res_precond = preconditioner(normal_operator)

                if isinstance(solver, IterativeLinearSolver):
                    inv_normal_op = solver(normal_operator, preconditioner=res_precond)
                else:
                    inv_normal_op = solver(normal_operator)

                # 3. Pre-compute cached vectors to accelerate linear/adjoint actions
                h_inv_u = inv_normal_op(model)
                denominator = self.model_space.inner_product(model, h_inv_u)

                residual = self.data_space.subtract(
                    self.forward_problem.forward_operator(model), data
                )
                r_inv_residual = (
                    self.forward_problem.data_error_measure.inverse_covariance(residual)
                )

                # Pre-compute w = (1/denom) * (L* u + (1/lambda) R^{-1} r)
                l_star_u = linear_part.adjoint(model)
                scaled_r_inv_res = self.data_space.multiply(
                    1.0 / damping, r_inv_residual
                )
                w_unscaled = self.data_space.add(l_star_u, scaled_r_inv_res)
                w = self.data_space.multiply(1.0 / denominator, w_unscaled)

                # 4. Forward derivative action
                def df_action(delta_data: Vector) -> Vector:
                    du_fixed = linear_part(delta_data)
                    delta_lambda = self.data_space.inner_product(w, delta_data)
                    return self.model_space.subtract(
                        du_fixed, self.model_space.multiply(delta_lambda, h_inv_u)
                    )

                # 5. Adjoint derivative action
                def df_adjoint(delta_model: Vector) -> Vector:
                    l_star_dy = linear_part.adjoint(delta_model)
                    scalar = self.model_space.inner_product(h_inv_u, delta_model)
                    return self.data_space.subtract(
                        l_star_dy, self.data_space.multiply(scalar, w)
                    )

                return LinearOperator(
                    self.data_space,
                    self.model_space,
                    df_action,
                    adjoint_mapping=df_adjoint,
                )

            return NonLinearOperator(
                self.data_space, self.model_space, mapping, derivative=derivative
            )

        else:
            forward_operator = self.forward_problem.forward_operator
            normal_operator = forward_operator @ forward_operator.adjoint
            inverse_normal_operator = solver(normal_operator)
            return forward_operator.adjoint @ inverse_normal_operator


class ConstrainedLinearMinimumNormInversion(LinearInversion):
    """
    Finds the minimum-norm solution subject to an affine subspace constraint.

    This class solves the regularized inverse problem using the discrepancy
    principle while strictly confining the solution to an affine subspace
    (e.g., enforcing a specific average property value or boundary condition).
    """

    def __init__(
        self, forward_problem: LinearForwardProblem, constraint: AffineSubspace
    ) -> None:
        """
        Initializes the constrained minimum norm inversion.

        This setup decomposes the problem into a base solution satisfying the
        affine constraint and a reduced unconstrained problem operating strictly
        within the constraint's parallel tangent space.

        Args:
            forward_problem: The original linear forward problem.
            constraint: The affine subspace defining the allowed model states.
        """
        super().__init__(forward_problem)
        if self.forward_problem.data_error_measure_set:
            self.assert_inverse_data_covariance()

        self._constraint = constraint

        proj_op = constraint.projection_operator
        self._u_base = proj_op.translation_part

        reduced_operator = forward_problem.forward_operator @ proj_op.linear_part

        self._reduced_forward_problem = LinearForwardProblem(
            reduced_operator,
            data_error_measure=(
                forward_problem.data_error_measure
                if forward_problem.data_error_measure_set
                else None
            ),
        )
        self._unconstrained_inversion = LinearMinimumNormInversion(
            self._reduced_forward_problem
        )

    def _get_reduced_operator(
        self,
        solver: LinearSolver,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
        significance_level: float = 0.95,
        minimum_damping: float = 0.0,
        maxiter: int = 100,
        rtol: float = 1.0e-6,
        atol: float = 0.0,
    ) -> NonLinearOperator:
        """
        Internal helper to construct the unconstrained reduced operator.

        Consolidates the instantiation of the underlying discrepancy principle
        search to ensure parameters are completely synchronized across both
        data-to-model and property-to-model mappings.
        """
        return self._unconstrained_inversion.minimum_norm_operator(
            solver,
            preconditioner=preconditioner,
            significance_level=significance_level,
            minimum_damping=minimum_damping,
            maxiter=maxiter,
            rtol=rtol,
            atol=atol,
        )

    def minimum_norm_operator(
        self,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
        significance_level: float = 0.95,
        minimum_damping: float = 0.0,
        maxiter: int = 100,
        rtol: float = 1.0e-6,
        atol: float = 0.0,
    ) -> NonLinearOperator:
        """
        Returns an operator that maps data to the constrained minimum-norm solution.
        The operator has its derivative set.

        Args:
            solver: The linear solver used for the normal equations.
            preconditioner: An optional preconditioner for iterative solvers.
            significance_level: The target probability mass for the chi-squared distribution.
            minimum_damping: The minimum allowed Tikhonov damping parameter.
            maxiter: Maximum iterations for the discrepancy principle root-finding.
            rtol: Relative tolerance for the damping parameter search.
            atol: Absolute tolerance for the damping parameter search.

        Returns:
            A `NonLinearOperator` mapping the data space to the model space.
        """
        reduced_op = self._get_reduced_operator(
            solver,
            preconditioner,
            significance_level,
            minimum_damping,
            maxiter,
            rtol,
            atol,
        )

        data_offset = self.forward_problem.forward_operator(self._u_base)
        domain = self.data_space
        codomain = self.model_space

        def mapping(d: Vector) -> Vector:
            d_tilde = domain.subtract(d, data_offset)
            w = reduced_op(d_tilde)
            return codomain.add(self._u_base, w)

        def derivative(d: Vector) -> LinearOperator:
            d_tilde = domain.subtract(d, data_offset)
            return reduced_op.derivative(d_tilde)

        return NonLinearOperator(domain, codomain, mapping, derivative=derivative)

    def constraint_value_mapping(
        self,
        data: Vector,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[Union[LinearOperator, LinearSolver]] = None,
        significance_level: float = 0.95,
        minimum_damping: float = 0.0,
        maxiter: int = 100,
        rtol: float = 1.0e-6,
        atol: float = 0.0,
    ) -> NonLinearOperator:
        r"""
        Returns an operator mapping a constraint value 'w' to the corresponding
        constrained minimum norm solution 'u' for a strictly fixed dataset.
        The operator has its derivative set.

        Args:
            data: The fixed observed data vector to fit.
            solver: The linear solver used for the normal equations.
            preconditioner: An optional preconditioner for iterative solvers.
            significance_level: The target probability mass for the chi-squared distribution.
            minimum_damping: The minimum allowed Tikhonov damping parameter.
            maxiter: Maximum iterations for the discrepancy principle root-finding.
            rtol: Relative tolerance for the damping parameter search.
            atol: Absolute tolerance for the damping parameter search.

        Returns:
            A `NonLinearOperator` mapping the constraint codomain (the property space)
            to the model space.

        Raises:
            ValueError: If the constraint subspace does not possess an explicit
                equation (i.e., the operator B is missing).
        """
        if not self._constraint.has_explicit_equation:
            raise ValueError(
                "Cannot compute mapping with respect to w: the constraint "
                "was defined geometrically without an explicit equation."
            )

        B_dagger = self._constraint.pseudo_inverse

        reduced_op = self._get_reduced_operator(
            solver,
            preconditioner,
            significance_level,
            minimum_damping,
            maxiter,
            rtol,
            atol,
        )

        A = self.forward_problem.forward_operator
        Id = self.model_space.identity_operator()

        domain = self._constraint.constraint_operator.codomain
        codomain = self.model_space

        def mapping(w: Vector) -> Vector:
            u_base = B_dagger(w)
            d_offset = A(u_base)
            d_tilde = self.data_space.subtract(data, d_offset)

            u_reduced = reduced_op(d_tilde)
            return self.model_space.add(u_base, u_reduced)

        def derivative(w: Vector) -> LinearOperator:
            u_base = B_dagger(w)
            d_offset = A(u_base)
            d_tilde = self.data_space.subtract(data, d_offset)

            D_unc = reduced_op.derivative(d_tilde)

            return (Id - D_unc @ A) @ B_dagger

        return NonLinearOperator(domain, codomain, mapping, derivative=derivative)



"""
Implements the Bayesian framework for solving linear inverse problems.

This module treats the inverse problem from a statistical perspective, aiming to
determine the full posterior probability distribution of the unknown model
parameters, rather than a single best-fit solution.

Key Classes
-----------
- `LinearBayesianInversion`: Computes the posterior Gaussian measure `p(u|d)`
  for the model `u` given observed data `d`.
- `ConstrainedLinearBayesianInversion`: Solves the inverse problem subject to
  an affine constraint `u in A`.
"""

from __future__ import annotations
from typing import Optional, List

from joblib import Parallel, delayed
import numpy as np

import scipy.sparse as sps
import scipy.sparse.linalg as splinalg

from .inversion import LinearInversion
from .gaussian_measure import GaussianMeasure
from .forward_problem import LinearForwardProblem
from .linear_operators import LinearOperator, DiagonalSparseMatrixLinearOperator
from .linear_solvers import LinearSolver, IterativeLinearSolver
from .hilbert_space import Vector, EuclideanSpace
from .subspaces import AffineSubspace


class LinearBayesianInversion(LinearInversion):
    """
    Solves a linear inverse problem using Bayesian methods.

    This class applies to problems of the form `d = A(u) + e`. It computes the
    full posterior probability distribution `p(u|d)`.
    """

    def __init__(
        self,
        forward_problem: LinearForwardProblem,
        model_prior_measure: GaussianMeasure,
        /,
    ) -> None:
        """
        Args:
            forward_problem: The forward problem linking the model to the data.
            model_prior_measure: The prior Gaussian measure on the model space.
        """
        super().__init__(forward_problem)
        self._model_prior_measure: GaussianMeasure = model_prior_measure

    @property
    def model_prior_measure(self) -> GaussianMeasure:
        """The prior Gaussian measure on the model space."""
        return self._model_prior_measure

    @property
    def normal_operator(self) -> LinearOperator:
        """
        Returns the Bayesian Normal operator: N = A Q A* + R.
        """
        forward_operator = self.forward_problem.forward_operator
        model_prior_covariance = self.model_prior_measure.covariance

        if self.forward_problem.data_error_measure_set:
            return (
                forward_operator @ model_prior_covariance @ forward_operator.adjoint
                + self.forward_problem.data_error_measure.covariance
            )
        else:
            return forward_operator @ model_prior_covariance @ forward_operator.adjoint

    def kalman_operator(
        self,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[LinearOperator] = None,
    ) -> LinearOperator:
        """
        Returns the Kalman gain operator K = Q A* N^-1.
        """
        forward_operator = self.forward_problem.forward_operator
        model_prior_covariance = self.model_prior_measure.covariance
        normal_operator = self.normal_operator

        if isinstance(solver, IterativeLinearSolver):
            inverse_normal_operator = solver(
                normal_operator, preconditioner=preconditioner
            )
        else:
            inverse_normal_operator = solver(normal_operator)

        return (
            model_prior_covariance @ forward_operator.adjoint @ inverse_normal_operator
        )

    def model_posterior_measure(
        self,
        data: Vector,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[LinearOperator] = None,
    ) -> GaussianMeasure:
        """
        Returns the posterior Gaussian measure p(u|d).

        Args:
            data: The observed data vector.
            solver: A linear solver for inverting the normal operator.
            preconditioner: An optional preconditioner.
        """
        data_space = self.data_space
        model_space = self.model_space
        forward_operator = self.forward_problem.forward_operator
        model_prior_covariance = self.model_prior_measure.covariance

        # 1. Compute Kalman Gain
        kalman_gain = self.kalman_operator(solver, preconditioner=preconditioner)

        # 2. Compute Posterior Mean
        # Shift data: d - A(mu_u)
        shifted_data = data_space.subtract(
            data, forward_operator(self.model_prior_measure.expectation)
        )

        # Shift for noise mean: d - A(mu_u) - mu_e
        if self.forward_problem.data_error_measure_set:
            error_expectation = self.forward_problem.data_error_measure.expectation
            shifted_data = data_space.subtract(shifted_data, error_expectation)
        else:
            error_expectation = data_space.zero

        mean_update = kalman_gain(shifted_data)
        expectation = model_space.add(self.model_prior_measure.expectation, mean_update)

        # 3. Compute Posterior Covariance (Implicitly)
        # C_post = C_u - K A C_u
        covariance = model_prior_covariance - (
            kalman_gain @ forward_operator @ model_prior_covariance
        )

        # 4. Set up Posterior Sampling
        # Logic: Can sample if prior is samplable AND (noise is absent OR samplable)
        can_sample_prior = self.model_prior_measure.sample_set
        can_sample_noise = (
            not self.forward_problem.data_error_measure_set
            or self.forward_problem.data_error_measure.sample_set
        )

        if can_sample_prior and can_sample_noise:

            def sample():
                # a. Sample Prior (u)
                model_sample = self.model_prior_measure.sample()

                # b. Calculate deterministic residual (v - Bu)
                prediction = forward_operator(model_sample)
                data_residual = data_space.subtract(data, prediction)

                # c. Subtract full noise sample (v - Bu - z)
                if self.forward_problem.data_error_measure_set:
                    noise_sample = self.forward_problem.data_error_measure.sample()
                    data_residual = data_space.subtract(data_residual, noise_sample)

                # d. Update with Kalman gain (u + K * residual)
                correction = kalman_gain(data_residual)
                return model_space.add(model_sample, correction)

            return GaussianMeasure(
                covariance=covariance, expectation=expectation, sample=sample
            )
        else:
            return GaussianMeasure(covariance=covariance, expectation=expectation)

    def diagonal_normal_preconditioner(
        self,
        /,
        *,
        blocks: Optional[List[List[int]]] = None,
        parallel: bool = False,
        n_jobs: int = -1,
    ) -> LinearOperator:
        """
        Constructs a diagonal preconditioner for the Bayesian normal operator
        (A Q A* + R) using an optimized formulation.

        This exploits the identity <v, A Q A* v> = <A* v, Q A* v>. If blocks
        of data indices are provided, it acts on the averaged basis vector
        for each block to compute a robust representative regional variance,
        requiring only one adjoint action of the forward operator per block.

        Args:
            blocks: An optional list of lists, where each sub-list contains indices
                of data points grouped together. Must perfectly partition the data space.
            parallel: If True, computes the adjoint actions in parallel.
            n_jobs: Number of parallel jobs to use. -1 means all available cores.

        Returns:
            A DiagonalSparseMatrixLinearOperator representing the inverse of
            the approximated normal operator.
        """

        data_space = self.data_space
        model_space = self.model_space
        A_adj = self.forward_problem.forward_operator.adjoint
        Q = self.model_prior_measure.covariance
        data_dim = data_space.dim

        if blocks is not None:
            flattened_indices = [idx for block in blocks for idx in block]

            if (
                len(flattened_indices) != data_dim
                or len(set(flattened_indices)) != data_dim
            ):
                raise ValueError(
                    f"The provided blocks must exactly partition the data space. "
                    f"Expected {data_dim} unique indices, but got {len(flattened_indices)} "
                    f"total indices with {len(set(flattened_indices))} unique."
                )

            if min(flattened_indices) < 0 or max(flattened_indices) >= data_dim:
                raise ValueError("Block indices are out of bounds for the data space.")

            blocks_to_compute = blocks
        else:
            blocks_to_compute = [[i] for i in range(data_dim)]

        def compute_block_aqa_diag(block: List[int]) -> float:
            """Worker function to compute the representative variance for a block."""
            n = len(block)

            # Form the averaged basis vector v = (1/n) * sum(e_i)
            c = np.zeros(data_dim)
            c[block] = 1.0 / n
            v = data_space.from_components(c)
            f = A_adj(v)
            Q_f = Q(f)
            return model_space.inner_product(f, Q_f)

        if parallel:
            computed_vals = Parallel(n_jobs=n_jobs)(
                delayed(compute_block_aqa_diag)(block) for block in blocks_to_compute
            )
        else:
            computed_vals = [
                compute_block_aqa_diag(block) for block in blocks_to_compute
            ]

        aqa_diag = np.zeros(data_dim)
        for block, val in zip(blocks_to_compute, computed_vals):
            aqa_diag[block] = val

        if self.forward_problem.data_error_measure_set:
            r_diag = (
                self.forward_problem.data_error_measure.covariance.extract_diagonal(
                    galerkin=True, parallel=parallel, n_jobs=n_jobs
                )
            )
            normal_diag = aqa_diag + r_diag
        else:
            normal_diag = aqa_diag

        approx_normal_op = DiagonalSparseMatrixLinearOperator.from_diagonal_values(
            data_space, data_space, normal_diag, galerkin=True
        )

        return approx_normal_op.inverse

    def sparse_localized_preconditioner(
        self,
        interacting_blocks: list[list[int]],
        rank: int = 10,
        parallel: bool = False,
        n_jobs: int = -1,
    ) -> LinearOperator:
        """
        Builds a sparse preconditioner for the Bayesian normal equations using
        randomized Nystrom approximations on localized, potentially overlapping sub-blocks.

        Args:
            interacting_blocks: A list of lists, where each sub-list contains the
                                indices of data points that strongly couple to each other.
            rank: The rank of the randomized Nystrom approximation to use per block.
            parallel: If True, computes the sub-block approximations in parallel.
            n_jobs: Number of CPU cores to use if parallel=True (-1 uses all cores).
        """

        forward_op = self.forward_problem.forward_operator
        prior_cov = self.model_prior_measure.covariance
        data_space = forward_op.codomain

        noise_variance = (
            self.forward_problem.data_error_measure.covariance.extract_diagonal(
                parallel=parallel, n_jobs=n_jobs, galerkin=True
            )
            if self.forward_problem.data_error_measure_set
            else np.zeros(self.data_space.dim)
        )

        core_normal_op = forward_op @ prior_cov @ forward_op.adjoint

        data_dim = data_space.dim
        euclidean_full = EuclideanSpace(data_dim)
        to_components_op = data_space.coordinate_projection

        print(
            f"Building sparse preconditioner over {len(interacting_blocks)} blocks (Rank {rank})..."
        )

        # 2. Define the worker function for a single block
        def _process_block(indices: list[int]):
            block_size = len(indices)
            idx_array = np.array(indices)

            # Restrict to this specific block of coordinates
            restrict_coords = euclidean_full.subspace_projection(indices)
            P = restrict_coords @ to_components_op
            P_star = P.adjoint

            # Localized Normal Operator: H_k = P * (A Q A*) * P*
            H_local = P @ core_normal_op @ P_star

            # Randomized Nystrom Approximation
            actual_rank = min(rank, block_size)

            # Unpack correctly: (eigenvectors, eigenvalues)
            evecs_op, evals_op = H_local.random_eig(actual_rank, method="fixed")

            # 1. Extract the dense eigenvector matrix natively (shape: block_size x actual_rank)
            V = evecs_op.matrix(dense=True, galerkin=True)

            # 2. Extract the 1D array of eigenvalues
            # Extracting the dense matrix of the diagonal operator and grabbing its diagonal
            Lambda = evals_op.matrix(dense=True).diagonal()

            # 3. Reconstruct the dense sub-block matrix purely in NumPy
            # (V * Lambda) efficiently broadcasts the 1D eigenvalues across the columns of V
            M_local_array = (V * Lambda) @ V.T

            # 4. Extract global coordinates for the sparse COO format
            I_local, J_local = np.meshgrid(
                np.arange(block_size), np.arange(block_size), indexing="ij"
            )
            I_global = idx_array[I_local.flatten()]
            J_global = idx_array[J_local.flatten()]

            V_flattened = M_local_array.flatten()

            return I_global, J_global, V_flattened

        # 3. Execute the block computations (Sequential or Parallel)
        row_indices, col_indices, values = [], [], []

        if parallel:
            results = Parallel(n_jobs=n_jobs)(
                delayed(_process_block)(indices) for indices in interacting_blocks
            )
            for I_glob, J_glob, V_flat in results:
                row_indices.extend(I_glob)
                col_indices.extend(J_glob)
                values.extend(V_flat)
        else:
            for indices in interacting_blocks:
                I_glob, J_glob, V_flat = _process_block(indices)
                row_indices.extend(I_glob)
                col_indices.extend(J_glob)
                values.extend(V_flat)

        # 4. Assemble the sparse normal matrix
        H_sparse = sps.coo_matrix(
            (values, (row_indices, col_indices)), shape=(data_dim, data_dim)
        )

        # 5. Add the data noise covariance (R) to the diagonal
        R_sparse = sps.diags(noise_variance)
        H_approx = (H_sparse + R_sparse).tocsc()

        # 6. Factorize using SuperLU
        print("Factorizing sparse matrix...")
        splu_solver = splinalg.splu(H_approx)

        # 7. Wrap the solver in a pygeoinf LinearOperator
        def apply_preconditioner(x):
            c = data_space.to_components(x)
            c_solved = splu_solver.solve(c)
            return data_space.from_components(c_solved)

        return LinearOperator(
            data_space,
            data_space,
            apply_preconditioner,
            adjoint_mapping=apply_preconditioner,
        )

    def surrogate_inversion(
        self,
        /,
        *,
        alternate_forward_operator: Optional[LinearOperator] = None,
        alternate_prior_measure: Optional[GaussianMeasure] = None,
        alternate_data_error_measure: Optional[GaussianMeasure] = None,
    ) -> LinearBayesianInversion:
        """
        Constructs a surrogate Bayesian inversion problem using simplified physics,
        priors, or data errors.

        This is primarily used to build robust, physics-based preconditioners.
        """

        A_tilde = alternate_forward_operator or self.forward_problem.forward_operator
        Q_tilde = alternate_prior_measure or self.model_prior_measure

        if alternate_data_error_measure is not None:
            R_tilde = alternate_data_error_measure
        elif self.forward_problem.data_error_measure_set:
            R_tilde = self.forward_problem.data_error_measure
        else:
            R_tilde = None

        if A_tilde.domain != Q_tilde.domain:
            raise ValueError(
                "The domain of the alternate forward operator must match "
                "the domain of the prior measure."
            )

        if A_tilde.codomain != self.data_space:
            raise ValueError(
                "The data space for the alternate forward operator must "
                "match that for the original problem"
            )

        if R_tilde.domain != self.data_space:
            raise ValueError(
                "The domain for the alternate error measure must "
                "match that for the original problem"
            )

        surrogate_forward_problem = LinearForwardProblem(
            A_tilde, data_error_measure=R_tilde
        )

        return LinearBayesianInversion(surrogate_forward_problem, Q_tilde)

    def surrogate_normal_preconditioner(
        self,
        solver: LinearSolver,
        /,
        *,
        alternate_forward_operator: Optional[LinearOperator] = None,
        alternate_prior_measure: Optional[GaussianMeasure] = None,
        alternate_data_error_measure: Optional[GaussianMeasure] = None,
    ) -> LinearOperator:
        """
        Builds a preconditioner by exactly inverting the normal operator of a
        simplified surrogate inverse problem.
        """
        # 1. Get the surrogate inverse problem
        surrogate_inv = self.surrogate_inversion(
            alternate_forward_operator=alternate_forward_operator,
            alternate_prior_measure=alternate_prior_measure,
            alternate_data_error_measure=alternate_data_error_measure,
        )

        # 2. Extract its normal operator and invert it using the provided solver
        return solver(surrogate_inv.normal_operator)


class ConstrainedLinearBayesianInversion(LinearInversion):
    """
    Solves a linear inverse problem subject to an affine subspace constraint.

    This class enforces the constraint `u in A` using either:
    1. Bayesian Conditioning (Default): p(u | d, u in A).
       If A is defined geometrically (no explicit equation), an implicit
       operator (I-P) is used, which requires a robust solver in the subspace.
    2. Geometric Projection: Projects the unconstrained posterior onto A.
    """

    def __init__(
        self,
        forward_problem: LinearForwardProblem,
        model_prior_measure: GaussianMeasure,
        constraint: AffineSubspace,
        /,
        *,
        geometric: bool = False,
    ) -> None:
        """
        Args:
            forward_problem: The forward problem.
            model_prior_measure: The unconstrained prior Gaussian measure.
            constraint: The affine subspace A.
            geometric: If True, uses orthogonal projection (Euclidean metric).
                       If False (default), uses Bayesian conditioning.
        """
        super().__init__(forward_problem)
        self._unconstrained_prior = model_prior_measure
        self._constraint = constraint
        self._geometric = geometric

    def conditioned_prior_measure(self) -> GaussianMeasure:
        """
        Computes the prior measure conditioned on the constraint.
        """
        return self._constraint.condition_gaussian_measure(
            self._unconstrained_prior, geometric=self._geometric
        )

    def model_posterior_measure(
        self,
        data: Vector,
        solver: LinearSolver,
        /,
        *,
        preconditioner: Optional[LinearOperator] = None,
    ) -> GaussianMeasure:
        """
        Returns the posterior Gaussian measure p(u | d, u in A).

        Args:
            data: Observed data vector.
            solver: Solver for the data update (inverts A C_cond A* + Ce).
            preconditioner: Preconditioner for the data update.

        Note: The solver for the constraint update is managed internally by
        the AffineSubspace object passed at initialization.
        """
        # 1. Condition Prior
        cond_prior = self.conditioned_prior_measure()

        # 2. Solve Bayesian Inverse Problem with the new prior
        bayes_inv = LinearBayesianInversion(self.forward_problem, cond_prior)

        return bayes_inv.model_posterior_measure(
            data, solver, preconditioner=preconditioner
        )

    def surrogate_inversion(
        self,
        /,
        *,
        alternate_forward_operator: Optional[LinearOperator] = None,
        alternate_prior_measure: Optional[GaussianMeasure] = None,
        alternate_data_error_measure: Optional[GaussianMeasure] = None,
        alternate_constraint: Optional[AffineSubspace] = None,
    ) -> ConstrainedLinearBayesianInversion:
        """
        Constructs a surrogate constrained Bayesian inversion problem using simplified
        physics, priors, data errors, or constraints.

        This is primarily used to build robust, physics-based preconditioners for
        constrained problems.
        """
        # 1. Substitute components or fall back to the exact ones
        A_tilde = alternate_forward_operator or self.forward_problem.forward_operator
        Q_tilde = alternate_prior_measure or self._unconstrained_prior

        if alternate_data_error_measure is not None:
            R_tilde = alternate_data_error_measure
        elif self.forward_problem.data_error_measure_set:
            R_tilde = self.forward_problem.data_error_measure
        else:
            R_tilde = None

        # Ensure domains match
        if A_tilde.domain != Q_tilde.domain:
            raise ValueError(
                "The domain of the alternate forward operator must match "
                "the domain of the prior measure."
            )

        if A_tilde.codomain != self.data_space:
            raise ValueError(
                "The data space for the alternate forward operator must "
                "match that for the original problem"
            )

        if R_tilde.domain != self.data_space:
            raise ValueError(
                "The domain for the alternate error measure must "
                "match that for the original problem"
            )

        # 2. Build the new surrogate forward problem
        surrogate_forward_problem = LinearForwardProblem(
            A_tilde, data_error_measure=R_tilde
        )

        # 3. Handle the constraint substitution
        C_tilde = alternate_constraint or self._constraint

        # 4. Return the new surrogate constrained inversion instance
        return ConstrainedLinearBayesianInversion(
            surrogate_forward_problem, Q_tilde, C_tilde, geometric=self._geometric
        )



from __future__ import annotations
from typing import Callable, Any, Optional, Tuple, List

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from pygeoinf.linear_operators import LinearOperator
from .symmetric_space import AbstractSymmetricLebesgueSpace, SymmetricSobolevSpace
from .circle import Lebesgue as CircleLebesgue
from .circle import Sobolev as CircleSobolev


class Lebesgue(AbstractSymmetricLebesgueSpace):
    """
    Implementation of the Lebesgue space L² on a line of interval.
    In the case of the line, it is assumed that the function has
    support limited to the chosen interval.

    A function is
    represented by its values on an evenly spaced grid. The co-ordinate basis for
    the space is through Fourier expansions. Details of the implementation are
    handled by mapping the function to one defined on a circle.
    """

    def __init__(self, kmax: int, /, *, a: float = 0.0, b: float = 1.0, c: float = 0.1):
        """
        Args:
        kmax: The maximum Fourier degree to be represented.
        a: The left endpoint of the interval. Defaults to 0.0
        b: The right endpoint of the interval. Dafaults to 0.1
        c: The padding distance for the computational domain.
               Defaults to 0.1
        """
        self._a = a
        self._b = b
        self._c = c

        length = b - a + 2 * c
        radius = length / (2 * np.pi)
        self._circle_space = CircleLebesgue(kmax, radius=radius)

        AbstractSymmetricLebesgueSpace.__init__(self, 1, kmax, 2 * kmax, False)

    @property
    def kmax(self) -> int:
        """
        Returns the maximum Fourier degree.
        """
        return self._circle_space.kmax

    @property
    def a(self) -> float:
        """
        Returns the left end point.
        """
        return self._a

    @property
    def b(self) -> float:
        """
        Returns the right end point.
        """
        return self._b

    @property
    def c(self) -> float:
        """
        Returns the padding distance.
        """
        return self._c

    @property
    def circle_space(self):
        """
        Returns the isomorphic space of functions on a circle.
        """
        return self._circle_space

    # ------------------------------------------------------ #
    #                     Public methods                     #
    # ------------------------------------------------------ #

    def points(self) -> np.ndarray:
        """Returns a numpy array of the grid points."""
        return np.fromiter(
            [self.angle_to_point(th) for th in self._circle_space.angles()],
            float,
        )

    def project_function(self, f: Callable[[float], float]) -> np.ndarray:
        """
        Returns an element of the space by projecting a given function.

        The function `f` is evaluated at the grid points of the space. To prevent
        spectral ringing (Gibbs phenomenon) from non-periodic functions, the
        function is smoothly tapered to zero within the padding regions
        [a-c, a] and [b, b+c] using a raised cosine window.

        Args:
            f: A function that takes a point (float) and returns a value.
        """
        points = self.points()

        vals = np.fromiter((f(x) for x in points), float)

        mask = np.ones_like(points)

        left_idx = (points < self.a) & (points >= self.a - self._c)
        if np.any(left_idx):
            x_norm_left = (points[left_idx] - (self.a - self._c)) / self._c
            mask[left_idx] = 0.5 * (1 - np.cos(np.pi * x_norm_left))

        right_idx = (points > self.b) & (points <= self.b + self._c)
        if np.any(right_idx):
            x_norm_right = (points[right_idx] - self.b) / self._c
            mask[right_idx] = 0.5 * (1 + np.cos(np.pi * x_norm_right))

        out_of_bounds = (points < self.a - self._c) | (points > self.b + self._c)
        mask[out_of_bounds] = 0.0

        return vals * mask

    def to_coefficients(self, u: np.ndarray) -> np.ndarray:
        """Maps a function vector to its complex Fourier coefficients."""
        return self._circle_space.to_coefficients(u)

    def from_coefficients(self, coeff: np.ndarray) -> np.ndarray:
        """Maps complex Fourier coefficients to a function vector."""
        return self._circle_space.from_coefficients(coeff)

    def point_to_angle(self, x: float) -> float:
        """Maps a point to the corresponding angle."""
        return (x - self.a + self._c) / self._circle_space.radius

    def angle_to_point(self, th: float) -> float:
        """Maps an angle to the corresponding point."""
        return self.a - self._c + self._circle_space.radius * th

    # ------------------------------------------------------ #
    #           Methods for SymmetricHilbertSpace            #
    # ------------------------------------------------------ #

    def integer_to_index(self, i: int) -> int:
        return self._circle_space.integer_to_index(i)

    def index_to_integer(self, k: int) -> int:
        return self._circle_space.index_to_integer(k)

    def laplacian_eigenvalue(self, k: int) -> float:
        return self._circle_space.laplacian_eigenvalue(k)

    def laplacian_eigenvector_squared_norm(self, k: int) -> float:
        return self._circle_space.laplacian_eigenvector_squared_norm(k)

    def laplacian_eigenvectors_at_point(self, x: float) -> np.ndarray:
        th = self.point_to_angle(x)
        return self._circle_space.laplacian_eigenvectors_at_point(th)

    def random_point(self) -> float:
        return np.random.uniform(self.a, self.b)

    def geodesic_distance(self, p1: float, p2: float) -> float:
        return abs(p2 - p1)

    def point_at_distance(self, p1: float, distance: float) -> float:
        """Returns a point separated from p1 by the given distance."""
        if p1 + distance <= self.b:
            return p1 + distance
        return p1 - distance

    def geodesic_quadrature(
        self, p1: float, p2: float, n_points: int
    ) -> Tuple[List[float], np.ndarray]:

        th1 = self.point_to_angle(p1)
        th2 = self.point_to_angle(p2)
        circle_points, weights = self._circle_space.geodesic_quadrature(
            th1, th2, n_points=n_points
        )

        points = [self.angle_to_point(th) for th in circle_points]

        return points, weights

    def with_degree(self, degree: int) -> Lebesgue:
        return Lebesgue(degree, a=self.a, b=self.b, c=self.c)

    def degree_transfer_operator(self, target_degree: int) -> LinearOperator:
        """
        Returns the transfer operator from this space to one with a different degree.
        """
        codomain = self.with_degree(target_degree)

        # We can directly leverage the isomorphic circle space's transfer operator
        circle_transfer = self._circle_space.degree_transfer_operator(target_degree)

        def mapping(u: np.ndarray) -> np.ndarray:
            return circle_transfer(u)

        def adjoint_mapping(v: np.ndarray) -> np.ndarray:
            return circle_transfer.adjoint(v)

        return LinearOperator(self, codomain, mapping, adjoint_mapping=adjoint_mapping)

    # ------------------------------------------------------ #
    #                 Methods for HilbertSpace               #
    # ------------------------------------------------------ #

    def to_components(self, x: np.ndarray) -> np.ndarray:
        return self._circle_space.to_components(x)

    def from_components(self, c: np.ndarray) -> np.ndarray:
        return self._circle_space.from_components(c)

    def is_element(self, x: Any) -> bool:
        return self._circle_space.is_element(x)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Lebesgue):
            return NotImplemented

        return (
            self.kmax == other.kmax
            and self.a == other.a
            and self.b == other.b
            and self._c == other._c
        )

    # ------------------------------------------------------ #
    #                 Methods for HilbertModule              #
    # ------------------------------------------------------ #

    def vector_multiply(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return self._circle_space.vector_multiply(x1, x2)

    def vector_sqrt(self, x: np.ndarray) -> np.ndarray:
        return self._circle_space.vector_sqrt(x)


class Sobolev(SymmetricSobolevSpace):
    """
    Implementation of the Sobolev space Hˢ on a circle.
    """

    def __init__(
        self,
        kmax: int,
        order: float,
        scale: float,
        /,
        *,
        a: float = 0.0,
        b: float = 1.0,
        c: float = None,
    ):
        """
        Args:
        kmax: The maximum Fourier degree to be represented.
        order: The Sobolev order, controlling the smoothness of functions.
        scale: The Sobolev length-scale.
        a: The left endpoint of the interval. Defaults to 0.0
        b: The right endpoint of the interval. Dafaults to 0.1
        c: The padding distance for the computational domain.
               Defaults to 6*scale
        """

        c = 6 * scale if c is None else c
        lebesgue_space = Lebesgue(kmax, a=a, b=b, c=c)
        SymmetricSobolevSpace.__init__(self, lebesgue_space, order, scale)

    @staticmethod
    def from_sobolev_parameters(
        order: float,
        scale: float,
        /,
        *,
        a: float = 0.0,
        b: float = 1.0,
        c: float = None,
        rtol: float = 1e-6,
        power_of_two: bool = False,
    ) -> "Sobolev":
        """
        Creates an instance with `kmax` chosen based on Sobolev parameters.

        The method estimates the truncation error for the Dirac measure and is
        only applicable for spaces with order > 0.5.

        Args:
            order: The Sobolev order. Must be > 0.5.
            scale: The Sobolev length-scale.
            a: The left endpoint of the interval. Defaults to 0.0
            b: The right endpoint of the interval. Dafaults to 0.1
            c: The padding distance for the computational domain.
               Defaults to 6*scale
            rtol: Relative tolerance used in assessing truncation error.
                Defaults to 1e-8.
            power_of_two: If True, `kmax` is set to the next power of two.

        Returns:
            An instance of the Sobolev class with an appropriate `kmax`.

        Raises:
            ValueError: If order is <= 0.5.
        """
        c = 6 * scale if c is None else c
        length = b - a + 2 * c
        radius = length / (2 * np.pi)
        circle_space = CircleSobolev.from_sobolev_parameters(
            order, scale, radius=radius, rtol=rtol, power_of_two=power_of_two
        )

        return Sobolev(circle_space.kmax, order, scale, a=a, b=b, c=c)

    # ---------------------------------------------- #
    #                   Properties                   #
    # -----------------------------------------------#

    @property
    def kmax(self) -> int:
        """
        Returns the maximum Fourier degree.
        """
        return self.underlying_space.kmax

    @property
    def a(self) -> float:
        """
        Returns the left end point.
        """
        return self.underlying_space.a

    @property
    def b(self) -> float:
        """
        Returns the right end point.
        """
        return self.underlying_space.b

    @property
    def c(self) -> float:
        """
        Returns the padding distance.
        """
        return self.underlying_space.c

    @property
    def circle_space(self):
        """
        Returns the isomorphic space of functions on a circle.
        """
        return CircleSobolev(
            self.kmax,
            self.order,
            self.scale,
            radius=self.underlying_space.circle_space.radius,
        )

    @property
    def derivative_operator(self):
        """
        Returns the derivative operator from the space to one with a lower order.
        """
        circle_op = self.circle_space.derivative_operator

        # Refactored: We no longer need to manually pass a, b, and c
        codomain = self.with_order(self.order - 1)

        # Calculate the chain rule scaling factor
        scaling = (2 * np.pi) / self.underlying_space.circle_space.radius

        # Scale the forward and adjoint mappings
        return LinearOperator(
            self,
            codomain,
            lambda u: scaling * circle_op(u),
            adjoint_mapping=lambda u: scaling * circle_op.adjoint(u),
        )

    # ---------------------------------------------- #
    #                 Public methods                 #
    # -----------------------------------------------#

    def with_order(self, order: float) -> Sobolev:
        return Sobolev(self.kmax, order, self.scale, a=self.a, b=self.b, c=self.c)

    def with_degree(self, degree: int) -> Sobolev:
        return Sobolev(degree, self.order, self.scale, a=self.a, b=self.b, c=self.c)

    def points(self) -> np.ndarray:
        """Returns a numpy array of the grid points."""
        return self.underlying_space.points()

    def project_function(self, f: Callable[float, float]) -> np.ndarray:
        """
        Returns an element of the space by projecting a given function.

        The function `f` is evaluated at the grid points of the space.

        Args:
            f: A function that takes an angle (float) and returns a value.
        """
        return self.underlying_space.project_function(f)

    def to_coefficients(self, u: np.ndarray) -> np.ndarray:
        """Maps a function vector to its complex Fourier coefficients."""
        return self.underlying_space.to_coefficients(u)

    def from_coefficients(self, coeff: np.ndarray) -> np.ndarray:
        """Maps complex Fourier coefficients to a function vector."""
        return self.underlying_space.from_coefficients(coeff)

    def point_to_angle(self, x: float) -> float:
        """
        Maps a point to the corresponding angle.
        """
        return self.underlying_space.point_to_angle(x)

    def angle_to_point(self, th: float) -> float:
        """
        Maps an angle to the corresponding point.
        """
        return self.underlying_space.angle_to_point(th)


def plot(
    space: Lebesgue | Sobolev,
    u: np.ndarray,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    full: Optional[bool] = False,
    **kwargs,
) -> Tuple[Figure, Axes]:
    """
    Makes a simple plot of a function on the circle.

    Args:
        space: The function space.
        u: The vector representing the function to be plotted.
        fig: An existing Matplotlib Figure object. Defaults to None.
        ax: An existing Matplotlib Axes object. Defaults to None.
        **kwargs: Keyword arguments forwarded to `ax.plot()`.

    Returns:
        A tuple (figure, axes) containing the plot objects.
    """
    figsize = kwargs.pop("figsize", (10, 8))

    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot()

    ax.plot(space.points(), u, **kwargs)
    if not full:
        ax.set_xlim(left=space.a, right=space.b)
    return fig, ax


def plot_error_bounds(
    space: Lebesgue | Sobolev,
    u: np.ndarray,
    u_bound: np.ndarray,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    full: Optional[bool] = False,
    **kwargs,
) -> Tuple[Figure, Axes]:
    """
    Plots a function on the circle along with its pointwise error bounds.

    This is particularly useful for visualizing Gaussian measures or Bayesian
    posterior uncertainties over the circular domain.

    Args:
        space: The function space.
        u: A 1D numpy array representing the mean function values.
        u_bound: A 1D numpy array giving the pointwise standard deviations or bounds.
        fig: An existing Matplotlib Figure object. If None, a new figure is created.
        ax: An existing Matplotlib Axes object. If None, a new subplot is added.
        **kwargs: Additional keyword arguments forwarded directly to `ax.fill_between()`
            (e.g., `alpha`, `color`).

    Returns:
        A tuple `(fig, ax)` containing the Matplotlib Figure and Axes objects.
    """
    figsize = kwargs.pop("figsize", (10, 8))

    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot()

    ax.fill_between(space.points(), u - u_bound, u + u_bound, **kwargs)
    if not full:
        ax.set_xlim(left=space.a, right=space.b)
    return fig, ax



from __future__ import annotations
from typing import Callable, Any, Optional, Tuple, List

import numpy as np
from scipy.fft import rfft, irfft

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from pygeoinf.linear_operators import LinearOperator
from .symmetric_space import AbstractSymmetricLebesgueSpace, SymmetricSobolevSpace


class Lebesgue(AbstractSymmetricLebesgueSpace):
    """
    Implementation of the Lebesgue space L² on a circle.

    This class represents square-integrable functions on a circle. A function is
    represented by its values on an evenly spaced grid. The co-ordinate basis for
    the space is through Fourier expansions.
    """

    def __init__(
        self,
        kmax: int,
        /,
        *,
        radius: float = 1.0,
    ):
        """
        Args:
        kmax: The maximum Fourier degree to be represented.
        radius: Radius of the circle. Defaults to 1.0.
        """

        if kmax <= 0:
            raise ValueError("kmax must be non-negative")

        if radius <= 0:
            raise ValueError("radius must be positive")

        self._kmax: int = kmax
        self._radius: float = radius

        self._fft_factor: float = np.sqrt(2 * np.pi * radius) / (2 * self.kmax)
        self._inverse_fft_factor: float = 1.0 / self._fft_factor

        AbstractSymmetricLebesgueSpace.__init__(self, 1, kmax, 2 * kmax, False)

    # ---------------------------------------------- #
    #                   Properties                   #
    # -----------------------------------------------#

    @property
    def kmax(self) -> int:
        """The maximum Fourier degree represented in this space."""
        return self._kmax

    @property
    def radius(self) -> float:
        """The radius of the circle."""
        return self._radius

    @property
    def angle_spacing(self) -> float:
        """The angular spacing between grid points."""
        return np.pi / self.kmax

    @property
    def fft_factor(self) -> float:
        """
        The factor by which the Fourier coefficients are scaled
        in forward transformations.
        """
        return self._fft_factor

    @property
    def inverse_fft_factor(self) -> float:
        """
        The factor by which the Fourier coefficients are scaled
        in inverse transformations.
        """
        return self._inverse_fft_factor

    # ---------------------------------------------- #
    #                 Public methods                 #
    # -----------------------------------------------#

    def angles(self) -> np.ndarray:
        """Returns a numpy array of the grid point angles."""
        return np.fromiter(
            [i * self.angle_spacing for i in range(2 * self.kmax)],
            float,
        )

    def project_function(self, f: Callable[[float], float]) -> np.ndarray:
        """
        Returns an element of the space by projecting a given function.

        The function `f` is evaluated at the grid points of the space.

        Args:
            f: A function that takes an angle (float) and returns a value.
        """
        return np.fromiter((f(theta) for theta in self.angles()), float)

    def to_coefficients(self, u: np.ndarray) -> np.ndarray:
        """Maps a function vector to its complex Fourier coefficients."""
        return rfft(u) * self.fft_factor

    def from_coefficients(self, coeff: np.ndarray) -> np.ndarray:
        """Maps complex Fourier coefficients to a function vector."""
        return irfft(coeff, n=2 * self.kmax) * self._inverse_fft_factor

    # ------------------------------------------------------ #
    #           Methods for SymmetricHilbertSpace            #
    # ------------------------------------------------------ #

    def integer_to_index(self, i: int) -> int:
        if i <= self.kmax:
            return i
        return -(i - self.kmax)

    def index_to_integer(self, k: int) -> int:
        if k >= 0:
            if k > self.kmax:
                raise ValueError(f"Index k={k} exceeds kmax={self.kmax}")
            return k

        i = abs(k) + self.kmax
        if i >= self.dim:
            raise ValueError(f"Index k={k} results in integer {i} out of bounds.")
        return i

    def laplacian_eigenvalue(self, k: int) -> float:
        return (k / self.radius) ** 2

    def laplacian_eigenvector_squared_norm(self, k: int) -> float:
        return 1.0 if k == 0 else 2.0

    def laplacian_eigenvectors_at_point(self, theta: float) -> np.ndarray:
        k_vals = np.arange(self.kmax + 1)
        cos_terms = np.cos(k_vals * theta)
        sin_terms = np.sin(k_vals[1 : self.kmax] * theta)
        return (
            self._metric
            @ np.concatenate([cos_terms, -sin_terms])
            / (np.sqrt(2 * np.pi * self.radius))
        )

    def random_point(self) -> float:
        return np.random.uniform(0, 2 * np.pi)

    def geodesic_distance(self, p1: float, p2: float) -> float:
        diff = (p2 - p1 + np.pi) % (2 * np.pi) - np.pi
        return float(np.abs(diff) * self.radius)

    def point_at_distance(self, p1: float, distance: float) -> float:
        return (p1 + distance / self.radius) % (2 * np.pi)

    def geodesic_quadrature(
        self, p1: float, p2: float, n_points: int
    ) -> Tuple[List[float], np.ndarray]:

        arc_length = self.geodesic_distance(p1, p2)

        diff = (p2 - p1 + np.pi) % (2 * np.pi) - np.pi

        x, w = np.polynomial.legendre.leggauss(n_points)

        t = (x + 1) / 2.0
        angles = p1 + t * diff

        scaled_weights = w * (arc_length / 2.0)

        return angles.tolist(), scaled_weights

    def with_degree(self, degree: int) -> Lebesgue:
        return Lebesgue(degree, radius=self.radius)

    def degree_transfer_operator(self, target_degree: int) -> LinearOperator:
        """
        Returns the transfer operator from this space to one with a different degree.

        This operator upsamples (by zero-padding Fourier coefficients) or
        downsamples (by truncating Fourier coefficients) the function grid.
        """
        codomain = self.with_degree(target_degree)

        def mapping(u: np.ndarray) -> np.ndarray:
            # 1. Move to the frequency domain
            c_in = self.to_coefficients(u)

            # 2. Pad or truncate
            c_out = np.zeros(target_degree + 1, dtype=complex)
            k_min = min(self.kmax, target_degree)
            c_out[: k_min + 1] = c_in[: k_min + 1]

            # 3. Enforce a strictly real Nyquist frequency when downsampling
            if target_degree < self.kmax:
                c_out[target_degree] = c_out[target_degree].real + 0j

            # 4. Return to the spatial domain
            return codomain.from_coefficients(c_out)

        def adjoint_mapping(v: np.ndarray) -> np.ndarray:
            c_in = codomain.to_coefficients(v)

            c_out = np.zeros(self.kmax + 1, dtype=complex)
            k_min = min(self.kmax, target_degree)
            c_out[: k_min + 1] = c_in[: k_min + 1]

            # The adjoint must mirror the forward map's Nyquist handling perfectly
            if self.kmax < target_degree:
                c_out[self.kmax] = c_out[self.kmax].real + 0j

            return self.from_coefficients(c_out)

        return LinearOperator(self, codomain, mapping, adjoint_mapping=adjoint_mapping)

    # ------------------------------------------------------ #
    #                 Methods for HilbertSpace               #
    # ------------------------------------------------------ #

    def to_components(self, x: np.ndarray) -> np.ndarray:
        coeff = self.to_coefficients(x)
        return self._coefficient_to_component(coeff)

    def from_components(self, c: np.ndarray) -> np.ndarray:
        coeff = self._component_to_coefficients(c)
        return self.from_coefficients(coeff)

    def is_element(self, x: Any) -> bool:
        if not isinstance(x, np.ndarray):
            return False
        if not x.shape == (self.dim,):
            return False
        return True

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Lebesgue):
            return NotImplemented

        return self.kmax == other.kmax and self.radius == other.radius

    # ------------------------------------------------------ #
    #                 Methods for HilbertModule              #
    # ------------------------------------------------------ #

    def vector_multiply(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return x1 * x2

    def vector_sqrt(self, x: np.ndarray) -> np.ndarray:
        return np.sqrt(x)

    # ------------------------------------------------------ #
    #                      Private methods                   #
    # ------------------------------------------------------ #

    def _coefficient_to_component(self, coeff: np.ndarray) -> np.ndarray:
        """Packs complex Fourier coefficients into a real component vector."""
        # For a real-valued input, the output of rfft (real FFT) has
        # conjugate symmetry. This implies that the imaginary parts of the
        # zero-frequency (k=0) and Nyquist-frequency (k=kmax) components
        # are always zero. We omit them from the component vector to create
        # a minimal, non-redundant representation.
        c = np.empty(self.dim, dtype=float)
        c[: self.kmax + 1] = coeff.real
        c[self.kmax + 1 :] = coeff.imag[1 : self.kmax]

        return c

    def _component_to_coefficients(self, c: np.ndarray) -> np.ndarray:
        """Unpacks a real component vector into complex Fourier coefficients."""
        # This is the inverse of `_coefficient_to_component`. It reconstructs
        # the full complex coefficient array that irfft expects. We re-insert
        # the known zeros for the imaginary parts of the zero-frequency (k=0)
        # and Nyquist-frequency (k=kmax) components, which were removed to
        # create the minimal real-valued representation.
        coeff = np.zeros(self.kmax + 1, dtype=complex)
        coeff.real = c[: self.kmax + 1]
        coeff.imag[1 : self.kmax] = c[self.kmax + 1 :]
        return coeff


class Sobolev(SymmetricSobolevSpace):
    """
    Implementation of the Sobolev space Hˢ on a circle.
    """

    def __init__(
        self,
        kmax: int,
        order: float,
        scale: float,
        /,
        *,
        radius: float = 1.0,
    ):
        """
        Args:
        kmax: The maximum Fourier degree to be represented.
        order: The Sobolev order, controlling the smoothness of functions.
        scale: The Sobolev length-scale.
        radius: Radius of the circle. Defaults to 1.0.
        """

        lebesgue_space = Lebesgue(kmax, radius=radius)
        SymmetricSobolevSpace.__init__(self, lebesgue_space, order, scale)

    @staticmethod
    def from_sobolev_parameters(
        order: float,
        scale: float,
        /,
        *,
        radius: float = 1.0,
        rtol: float = 1e-6,
        power_of_two: bool = False,
    ) -> "Sobolev":
        """
        Creates an instance with `kmax` chosen based on Sobolev parameters.

        The method estimates the truncation error for the Dirac measure and is
        only applicable for spaces with order > 0.5.

        Args:
            order: The Sobolev order. Must be > 0.5.
            scale: The Sobolev length-scale.
            radius: The radius of the circle. Defaults to 1.0.
            rtol: Relative tolerance used in assessing truncation error.
                Defaults to 1e-8.
            power_of_two: If True, `kmax` is set to the next power of two.

        Returns:
            An instance of the Sobolev class with an appropriate `kmax`.

        Raises:
            ValueError: If order is <= 0.5.
        """
        if order <= 0.5:
            raise ValueError("This method is only applicable for orders > 0.5")

        summation = 1.0
        k = 0
        err = 1.0
        while err > rtol:
            k += 1
            term = (1 + (scale * k / radius) ** 2) ** -order
            summation += 2 * term
            err = 2 * term / summation
            if k > 100000:
                raise RuntimeError("Failed to converge on a stable kmax.")

        if power_of_two:
            n = int(np.log2(k))
            k = 2 ** (n + 1)

        return Sobolev(k, order, scale, radius=radius)

    # ---------------------------------------------- #
    #                   Properties                   #
    # -----------------------------------------------#

    @property
    def kmax(self) -> int:
        """The maximum Fourier degree represented in this space."""
        return self.underlying_space.kmax

    @property
    def radius(self) -> float:
        """The radius of the circle."""
        return self.underlying_space.radius

    @property
    def angle_spacing(self) -> float:
        """The angular spacing between grid points."""
        return self.underlying_space.angle_spacing

    @property
    def fft_factor(self) -> float:
        """
        The factor by which the Fourier coefficients are scaled
        in forward transformations.
        """
        return self.underlying_space.fft_factor

    @property
    def inverse_fft_factor(self) -> float:
        """
        The factor by which the Fourier coefficients are scaled
        in inverse transformations.
        """
        return self.underlying_space.inverse_fft_factor

    @property
    def derivative_operator(self) -> LinearOperator:
        """
        Returns the derivative operator from the space to one with a lower order.
        """
        codomain = self.with_order(self.order - 1)

        lebesgue_space = self.underlying_space
        k = np.arange(self.kmax + 1)

        def mapping(u):
            coeff = lebesgue_space.to_coefficients(u)
            diff_coeff = 1j * k * coeff
            return lebesgue_space.from_coefficients(diff_coeff)

        op_L2 = LinearOperator(
            lebesgue_space,
            lebesgue_space,
            mapping,
            adjoint_mapping=lambda u: -1 * mapping(u),
        )

        return LinearOperator.from_formal_adjoint(self, codomain, op_L2)

    # ---------------------------------------------- #
    #                 Public methods                 #
    # -----------------------------------------------#

    def with_order(self, order: float) -> Sobolev:
        return Sobolev(self.kmax, order, self.scale, radius=self.radius)

    def with_degree(self, degree: int) -> Sobolev:
        return Sobolev(degree, self.order, self.scale, radius=self.radius)

    def angles(self) -> np.ndarray:
        """Returns a numpy array of the grid point angles."""
        return self.underlying_space.angles()

    def project_function(self, f: Callable[[float], float]) -> np.ndarray:
        """
        Returns an element of the space by projecting a given function.

        The function `f` is evaluated at the grid points of the space.

        Args:
            f: A function that takes an angle (float) and returns a value.
        """
        return self.underlying_space.project_function(f)

    def to_coefficients(self, u: np.ndarray) -> np.ndarray:
        """Maps a function vector to its complex Fourier coefficients."""
        return self.underlying_space.to_coefficients(u)

    def from_coefficients(self, coeff: np.ndarray) -> np.ndarray:
        """Maps complex Fourier coefficients to a function vector."""
        return self.underlying_space.from_coefficients(coeff)


# ------------------------------------------------- #
#           Associated plotting functions           #
# ------------------------------------------------- #


def plot(
    space: Lebesgue | Sobolev,
    u: np.ndarray,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Tuple[Figure, Axes]:
    """
    Creates a simple line plot of a function on the circle.

    Args:
        angles: The function space.
        u: A 1D numpy array representing the function values (the y-axis).
        fig: An existing Matplotlib Figure object. If None, a new figure is created.
        ax: An existing Matplotlib Axes object. If None, a new subplot is added.
        **kwargs: Additional keyword arguments forwarded directly to `ax.plot()`
            (e.g., `color`, `linewidth`, `label`).

    Returns:
        A tuple `(fig, ax)` containing the Matplotlib Figure and Axes objects.
    """
    figsize = kwargs.pop("figsize", (10, 8))

    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot()

    ax.plot(space.angles(), u, **kwargs)
    return fig, ax


def plot_error_bounds(
    space: Lebesgue | Sobolev,
    u: np.ndarray,
    u_bound: np.ndarray,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Tuple[Figure, Axes]:
    """
    Plots a function on the circle along with its pointwise error bounds.

    This is particularly useful for visualizing Gaussian measures or Bayesian
    posterior uncertainties over the circular domain.

    Args:
        space: The function space.
        u: A 1D numpy array representing the mean function values.
        u_bound: A 1D numpy array giving the pointwise standard deviations or bounds.
        fig: An existing Matplotlib Figure object. If None, a new figure is created.
        ax: An existing Matplotlib Axes object. If None, a new subplot is added.
        **kwargs: Additional keyword arguments forwarded directly to `ax.fill_between()`
            (e.g., `alpha`, `color`).

    Returns:
        A tuple `(fig, ax)` containing the Matplotlib Figure and Axes objects.
    """
    figsize = kwargs.pop("figsize", (10, 8))

    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot()

    ax.fill_between(space.angles(), u - u_bound, u + u_bound, **kwargs)
    return fig, ax
